%% Modified_LSM_Algorithm
% This Matlab code performs the modified LSM algorithm in Hong et al. (2018)
% This is the replicated model of the modified LSM algorithm.

% 1. Monte Carlo Sampling
% This section initializes the necessary input parameters to perform the
% Monte Carlo sampling of the model parameters of the two-factor production model 

N_MC = 100;
FieldLifeTime = 50;

OOIP_Mean = 240; 
E_inf_Phase1_Mean = 0.2; 
Tau_Phase1_Mean = 16; 
dE_inf_Phase2_Mean = 0.15; 
Tau_Phase2_Mean = 7; 

OOIP_SD = 35; 
E_inf_Phase1_SD = 0.05; 
Tau_Phase1_SD = 2; 
dE_inf_Phase2_SD = 0.05; 
Tau_Phase2_SD = 1.5; 

ErrorMean4Rate_Pct = 0;
ErrorSD4Rate_Pct = 0.10;

% Initialization of parameters for multivariate normal distribution 
mu = [OOIP_Mean, E_inf_Phase1_Mean, Tau_Phase1_Mean, dE_inf_Phase2_Mean, Tau_Phase2_Mean];
ExpSigma = [OOIP_SD, E_inf_Phase1_SD, Tau_Phase1_SD, dE_inf_Phase2_SD, Tau_Phase2_SD];
ExpCorrC = [1 -0.8 0.16 0.56 -0.08; -0.8 1 0.2 -0.7 0.1; 0.16 0.2 1 -0.3 -0.2; 0.56 -0.7 -0.3 1 -0.3; -0.08 0.1 -0.2 -0.3 1];
cov = ExpCorrC.*(ExpSigma'*ExpSigma);
PriorRealizations = mvnrnd(mu, cov, N_MC);

% Prior realizations of the model parameters 
OOIP_prior = PriorRealizations(:,1); OOIP_prior(OOIP_prior<10) = 10; OOIP_prior(OOIP_prior>1000) = 1000;
E_inf_Phase1_prior = PriorRealizations(:,2); E_inf_Phase1_prior(E_inf_Phase1_prior<0.05) = 0.05; E_inf_Phase1_prior(E_inf_Phase1_prior>0.5) = 0.5;
Tau_Phase1_prior = PriorRealizations(:,3); Tau_Phase1_prior(Tau_Phase1_prior<1) = 1; Tau_Phase1_prior(Tau_Phase1_prior>30) = 30;
dE_inf_Phase2_prior = PriorRealizations(:,4); dE_inf_Phase2_prior(dE_inf_Phase2_prior<0.01) = 0.01; dE_inf_Phase2_prior(dE_inf_Phase2_prior>0.31) = 0.31;
Tau_Phase2_prior = PriorRealizations(:,5); Tau_Phase2_prior(Tau_Phase2_prior<1) = 1; Tau_Phase2_prior(Tau_Phase2_prior>13) = 13;


%  2. Two-Factor Production Model 
%  This section performs the calculation of the recovery factors and the respective modeled oil production rates

Recovery_Phase1_real_time = zeros(N_MC,FieldLifeTime);
Recovery_Phase1_real_EndTime = zeros(N_MC,1);
Recovery_Phase2_real_time = zeros(N_MC,1);
Recovery_Phase1n2_real_time = zeros(N_MC,FieldLifeTime);
Recovery_Phase1n2_real_time_ShiftTime = zeros(N_MC,FieldLifeTime,FieldLifeTime+1);
Rate_Phase1n2_real_time_ShiftTime = zeros(N_MC,FieldLifeTime,FieldLifeTime+1);

for i_MC = 1: N_MC
    for time = 1: FieldLifeTime %this is the time shift in rows
    Recovery_Phase1_real_time(i_MC,time) = (E_inf_Phase1_prior(i_MC)).*(1-exp(-time/Tau_Phase1_prior(i_MC))); %computation of the E_R1
    if time == 1
        Recovery_Phase1_real_EndTime(i_MC) = 0;  
    else 
        Recovery_Phase1_real_EndTime(i_MC) = Recovery_Phase1_real_time(i_MC,time-1);
    end
    
    Recovery_Phase2_real_time(i_MC) = Recovery_Phase1_real_EndTime(i_MC) + dE_inf_Phase2_prior(i_MC).*(1-exp(-1/Tau_Phase2_prior(i_MC)));
    
    for t = 1: FieldLifeTime %this is the time shift in columns
      %computation of the E_R with both R1 and R2
      if t < time
      Recovery_Phase1n2_real_time(i_MC,t) = Recovery_Phase1_real_time(i_MC,t); 
      elseif t == time
      Recovery_Phase1n2_real_time(i_MC,t) = Recovery_Phase2_real_time(i_MC);
      else
      Recovery_Phase1n2_real_time(i_MC,t) = Recovery_Phase1_real_EndTime(i_MC) + dE_inf_Phase2_prior(i_MC).*(1-exp(-(t-time+1)/Tau_Phase2_prior(i_MC)));    
      end
      Recovery_Phase1n2_real_time_ShiftTime(i_MC,t,time) = Recovery_Phase1n2_real_time(i_MC,t);
      Recovery_Phase1n2_real_time_ShiftTime(i_MC,t,FieldLifeTime+1) = Recovery_Phase1_real_time(i_MC,t);
    
      if t == 1 %computation of the flow rate at the first year
      Rate_Phase1n2_real_time_ShiftTime(i_MC,t,time) = OOIP_prior(i_MC).*Recovery_Phase1n2_real_time_ShiftTime(i_MC,t,time);
      Rate_Phase1n2_real_time_ShiftTime(i_MC,t,FieldLifeTime+1) = OOIP_prior(i_MC).*Recovery_Phase1n2_real_time_ShiftTime(i_MC,t,FieldLifeTime+1);
      else %computation of the flow rate at the following years
      Rate_Phase1n2_real_time_ShiftTime(i_MC,t,time) = OOIP_prior(i_MC).*(Recovery_Phase1n2_real_time_ShiftTime(i_MC,t,time)-Recovery_Phase1n2_real_time_ShiftTime(i_MC,t-1,time)); 
      Rate_Phase1n2_real_time_ShiftTime(i_MC,t,FieldLifeTime+1) = OOIP_prior(i_MC).*(Recovery_Phase1n2_real_time_ShiftTime(i_MC,t,FieldLifeTime+1)-Recovery_Phase1n2_real_time_ShiftTime(i_MC,t-1,FieldLifeTime+1));  
      end
       
    end
    end
end

% Computation of the measurement error and the measured oil production rates
ErrorSD_Matrix = ErrorSD4Rate_Pct.*Rate_Phase1n2_real_time_ShiftTime;
ObsRate_Phase1n2_real_time_ShiftTime = normrnd(Rate_Phase1n2_real_time_ShiftTime,ErrorSD_Matrix);


% 3. Calculation of Cashflow and NPV
% This section computes te cashflow and the NPV used in the example problem

N_MC = size(ObsRate_Phase1n2_real_time_ShiftTime,1);
FieldLifeTime = size(ObsRate_Phase1n2_real_time_ShiftTime,2);
CashFlow = zeros(N_MC,FieldLifeTime);
DisCashFlow = zeros(N_MC,FieldLifeTime);
DisCashFlow_ShiftTime = zeros(N_MC,FieldLifeTime,FieldLifeTime+1);
NPV_reals = zeros(N_MC,1);
NPVtable_LTPhase1_LTPhase2_real = zeros(FieldLifeTime+1,FieldLifeTime+1,N_MC);

% Values of Economic Parameters
OilPrice = 50; 
Capex_Phase1 = 50; 
Capex_Phase2After1 = 40;
Capex_Phase2No1 = 75; 
Opex_Phase1 = 20; 
Opex_Phase2 = 30;
DisRate = 0.12;       

for t = 1: FieldLifeTime
    for time = 1: FieldLifeTime+1
        if t == time 
            if time == 1
                CashFlow(:,t) = OilPrice.*ObsRate_Phase1n2_real_time_ShiftTime(:,t,time) - Opex_Phase2 - Capex_Phase2No1;
                DisCashFlow(:,t) = CashFlow(:,t)/((1+DisRate).^(t)); 
            else
               CashFlow(:,t) = OilPrice.*ObsRate_Phase1n2_real_time_ShiftTime(:,t,time) - Opex_Phase2 - Capex_Phase2After1;
               DisCashFlow(:,t) = CashFlow(:,t)/((1+DisRate).^(t));  
            end
        elseif t < time
            if t == 1
                CashFlow(:,t) = OilPrice.*ObsRate_Phase1n2_real_time_ShiftTime(:,t,time) - Opex_Phase1 - Capex_Phase1;
                DisCashFlow(:,t) = CashFlow(:,t)/((1+DisRate).^(t)); 
            else
                CashFlow(:,t) = OilPrice.*ObsRate_Phase1n2_real_time_ShiftTime(:,t,time) - Opex_Phase1;
                DisCashFlow(:,t) = CashFlow(:,t)/((1+DisRate).^(t)); 
            end
        else
            CashFlow(:,t) = OilPrice.*ObsRate_Phase1n2_real_time_ShiftTime(:,t,time) - Opex_Phase2;
            DisCashFlow(:,t) = CashFlow(:,t)/((1+DisRate).^(t)); 
        end    
        
        DisCashFlow_ShiftTime(:,t,time) = DisCashFlow(:,t);
        
        if time == 1
                NPV_reals(:) = sum(DisCashFlow_ShiftTime(:,1:t,time),2);
                NPVtable_LTPhase1_LTPhase2_real(time,t+1,:) = NPV_reals(:);
        elseif time == FieldLifeTime+1
                NPV_reals(:) = sum(DisCashFlow_ShiftTime(:,1:FieldLifeTime,time),2);
                NPVtable_LTPhase1_LTPhase2_real(time,1,:) = NPV_reals(:);
        else
            if t == time 
                NPV_reals(:) = sum(DisCashFlow_ShiftTime(:,1:time,time),2);
                NPVtable_LTPhase1_LTPhase2_real(time,2,:) = NPV_reals(:);
            elseif t < time
                NPV_reals(:) = sum(DisCashFlow_ShiftTime(:,1:time-1,time),2);
                NPVtable_LTPhase1_LTPhase2_real(time,1,:) = NPV_reals(:);
            else
                NPV_reals(:) = sum(DisCashFlow_ShiftTime(:,1:t,time),2);
                NPVtable_LTPhase1_LTPhase2_real(time,t-time+2,:) = NPV_reals(:);
            end
        end
    end
end


% 4. Determination of DWOI, VOWI, DWPI, and VOPI
% This section determines the Decision Without Information and Decision With Perfect Information
% The corresponding Value Without Information and Value of Perfect Information are estimated

N_MC = size(ObsRate_Phase1n2_real_time_ShiftTime,1);
FieldLifeTime = size(ObsRate_Phase1n2_real_time_ShiftTime,2);
N_Phase1n2LifeTime = sum(1:FieldLifeTime+1);
NPVvector_1real = zeros(N_Phase1n2LifeTime,3);
NPVmatrix_reals = zeros(N_Phase1n2LifeTime,N_MC);
Sum_NPVvector_1real = zeros(N_Phase1n2LifeTime,1);
meanNPVvector = zeros(N_Phase1n2LifeTime,3);
NPV_element_vector = zeros(FieldLifeTime+1,FieldLifeTime+1);
Phase1n2LifeTimeTable = zeros(N_Phase1n2LifeTime,2);

VWPI_real = zeros(N_MC,1);
DWPI_Phase1LifeTime_real = zeros(N_MC,1);
DWPI_Phase2LifeTime_real = zeros(N_MC,1);

for t = 1: FieldLifeTime
    for time = 1: FieldLifeTime+1
        NPV_element_vector(time,t+1) = t;
        if time > 1
            NPV_element_vector(time,FieldLifeTime-time+3:FieldLifeTime+1) = FieldLifeTime+1;
        else
            NPV_element_vector(time,t+1) = t;    
        end
    end
end

[row_NPV_element,col_NPV_element]=find(NPV_element_vector(:,:) < FieldLifeTime+1);

for k = 1: N_Phase1n2LifeTime
    NPVvector_1real(k,1) = col_NPV_element(k)-1;
    NPVvector_1real(k,2) = row_NPV_element(k)-1;
    meanNPVvector(k,1) = col_NPV_element(k)-1;
    meanNPVvector(k,2) = row_NPV_element(k)-1;
end
     
for i_MC = 1: N_MC
    for k = 1: N_Phase1n2LifeTime 
        NPVvector_1real(k,3) = NPVtable_LTPhase1_LTPhase2_real(col_NPV_element(k),row_NPV_element(k),i_MC);
        NPVmatrix_reals(k,i_MC) = NPVvector_1real(k,3);
        [VWPI_real(i_MC),DWPI_idx] = max(NPVvector_1real(:,3));
        DWPI_Phase1LifeTime_real(i_MC) = NPVvector_1real(DWPI_idx,1);
        DWPI_Phase2LifeTime_real(i_MC) = NPVvector_1real(DWPI_idx,2);
    end
end

meanNPV = mean(NPVmatrix_reals,2);

for k = 1: N_Phase1n2LifeTime 
    meanNPVvector(k,3) = meanNPV(k,1);
end

[EVWOI,DWOI_idx] = max(meanNPVvector(:,3));
DWOI_Phase1LifeTime = col_NPV_element(DWOI_idx)-1; 
DWOI_Phase2LifeTime = row_NPV_element(DWOI_idx)-1;
DWOI_LifeTime = DWOI_Phase1LifeTime + DWOI_Phase2LifeTime;
DWPI_LifeTime_real = DWPI_Phase1LifeTime_real + DWPI_Phase2LifeTime_real;
Phase1n2LifeTimeTable(:,1) = meanNPVvector(:,1);
Phase1n2LifeTimeTable(:,2) = meanNPVvector(:,2);
EVWPI = mean(VWPI_real);
VOPI = EVWPI - EVWOI;

%  5. Path Table Generation (Modified LSM algorithm)
%  This section determines the optimal stopping time given the switch time at every year
%  Regression Analysis is applied
%  The NPV corresponding to the optimal stopping time is recorded into the PathTable

N_MC = size(ObsRate_Phase1n2_real_time_ShiftTime,1);
FieldLifeTime = size(ObsRate_Phase1n2_real_time_ShiftTime,2);
SRDM_Shift_Stop_OptNPV_real = zeros(N_MC,1);
        
PathTable = zeros(N_MC,FieldLifeTime+1);
PathTable(:,end) = NPVtable_LTPhase1_LTPhase2_real(FieldLifeTime+1,1,:);
Phase2_StopTime_real_Phase1LifeTime = zeros(N_MC,FieldLifeTime+1);

for k_Shift_time = 1: FieldLifeTime
        NPV_2Phases_matrix = zeros(N_MC,k_Shift_time+1); 
    for k_Stop_time = 1: k_Shift_time+1          
        NPV_2Phases_matrix(:,k_Stop_time) = NPVtable_LTPhase1_LTPhase2_real(FieldLifeTime-k_Shift_time+1,k_Stop_time,:);
    end
    for k_StopTime = 1: k_Shift_time    
        if k_StopTime == 1 
            x_data_stop_SRDM = ObsRate_Phase1n2_real_time_ShiftTime(:,1:FieldLifeTime-k_StopTime,FieldLifeTime-k_Shift_time+1); %%%%%
            X_stop_SRDM = [ones(N_MC,1) x_data_stop_SRDM]; 
            y_data_stop_matrix = [NPV_2Phases_matrix(:,k_Shift_time-k_StopTime+1) NPV_2Phases_matrix(:,k_Shift_time-k_StopTime+2)];
            [Stop_value,Stop_idx] = max([X_stop_SRDM*regress(NPV_2Phases_matrix(:,k_Shift_time-k_StopTime+1),X_stop_SRDM) X_stop_SRDM*regress(NPV_2Phases_matrix(:,k_Shift_time-k_StopTime+2),X_stop_SRDM)],[],2);
            for i_MC = 1: N_MC
                SRDM_Shift_Stop_OptNPV_real(i_MC,1) = y_data_stop_matrix(i_MC,Stop_idx(i_MC));
            end    
        else
            x_data_stop_SRDM = ObsRate_Phase1n2_real_time_ShiftTime(:,1:FieldLifeTime-k_StopTime,FieldLifeTime-k_Shift_time+1); %%%%%
            X_stop_SRDM = [ones(N_MC,1) x_data_stop_SRDM]; 
            y_data_stop_matrix = [NPV_2Phases_matrix(:,k_Shift_time-k_StopTime+1) SRDM_Shift_Stop_OptNPV_real(:,1)];
            [Stop_value,Stop_idx] = max([X_stop_SRDM*regress(NPV_2Phases_matrix(:,k_Shift_time-k_StopTime+1),X_stop_SRDM) X_stop_SRDM*regress(SRDM_Shift_Stop_OptNPV_real(:,1),X_stop_SRDM)],[],2);
            for i_MC = 1: N_MC
                SRDM_Shift_Stop_OptNPV_real(i_MC,1) = y_data_stop_matrix(i_MC,Stop_idx(i_MC));
            end  
        end
    end     
    for i_MC = 1: N_MC
        PathTable(i_MC,FieldLifeTime-k_Shift_time+1) = SRDM_Shift_Stop_OptNPV_real(i_MC,1);
        for k_Stop_time = 1: k_Shift_time+1
            if SRDM_Shift_Stop_OptNPV_real(i_MC,1) == NPV_2Phases_matrix(i_MC,k_Stop_time)
                Phase2_StopTime_real_Phase1LifeTime(i_MC,FieldLifeTime-k_Shift_time+1) = k_Stop_time-1;
            end
        end
    end  
end

% 6. SRDM Approach (Modified LSM algorithm)
% This section determines the optimal decision policy based on SRDM approach

FieldLifeTime = size(ObsRate_Phase1n2_real_time_ShiftTime,2);
y_data_continuation = zeros(N_MC,1);
y_data_shift = zeros(N_MC,1);
y_reg_continuation = zeros(N_MC,1);
y_reg_shift = zeros(N_MC,1);
x_data = ObsRate_Phase1n2_real_time_ShiftTime(:,:,FieldLifeTime+1);

ValueTable = zeros(N_MC,FieldLifeTime+1);
ValueTable(:,end) = PathTable(:,end);
SRDM_DecisionTable = zeros(N_MC,FieldLifeTime+1);
SRDM_DecisionTable(:,end) = 1;

for k_Shift_time = 1:FieldLifeTime
    X = [ones(N_MC,1) x_data(:,1:FieldLifeTime-k_Shift_time)]; %%%%%
    y_data_shift(:) = PathTable(:,FieldLifeTime-k_Shift_time+1); 
    coef_shift = regress(y_data_shift,X);
    y_reg_shift(:) = X*coef_shift;
    y_data_continuation(:) = ValueTable(:,FieldLifeTime-k_Shift_time+2);
    coef_continuation = regress(y_data_continuation,X);
    y_reg_continuation(:) = X*coef_continuation;
    ENPVTable_Comparison = [y_reg_shift, y_reg_continuation];
    NPV_SRDM = [y_data_shift, y_data_continuation];
    [ENPV_SRDM,ENPV_SRDM_idx] = max(ENPVTable_Comparison,[],2);
    
    for i_MC = 1: N_MC
        ValueTable(i_MC,FieldLifeTime-k_Shift_time+1) = NPV_SRDM(i_MC,ENPV_SRDM_idx(i_MC));
        if y_reg_continuation(i_MC) > y_reg_shift(i_MC)
            SRDM_DecisionTable(i_MC,FieldLifeTime-k_Shift_time+1) = 0;
        else
            SRDM_DecisionTable(i_MC,FieldLifeTime-k_Shift_time+1) = 1;
        end
    end
end

ShiftValue = PathTable(:,1);
ExpShiftValue = mean(ShiftValue);

ContinuationValue = ValueTable(:,1);
ExpContinuationValue = mean(ContinuationValue);
EVWII = ExpContinuationValue;
VOI = EVWII - EVWOI;
VOI_over_EVWOI_pct = VOI/EVWOI*100;

[~,DWI_Phase1LifeTime_idx] = max(SRDM_DecisionTable,[],2);
DWI_Phase1LifeTime = DWI_Phase1LifeTime_idx-1;

DWI_Phase2LifeTime = zeros(N_MC,1);
for i_MC = 1:N_MC
    Phase1LifeTime_idx_real = DWI_Phase1LifeTime_idx(i_MC,1);
    DWI_Phase2LifeTime(i_MC,1) = Phase2_StopTime_real_Phase1LifeTime(i_MC,Phase1LifeTime_idx_real);
end

DWI_LifeTime = DWI_Phase1LifeTime+DWI_Phase2LifeTime;


 





